@@ -1,8 +1,8 @@
 //
 // anyRemote
-// a bluetooth remote for your PC.
+// a wi-fi or bluetooth remote for your PC.
 //
-// Copyright (C) 2006-2013 Mikhail Fedotov <anyremote@mail.ru>
+// Copyright (C) 2006-2014 Mikhail Fedotov <anyremote@mail.ru>
 //
 // This program is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -38,7 +38,8 @@
 #define IS_OK(answer) (strstr(answer, "OK") != NULL)
 
 //#define DEFAULT_DEVICE  "/dev/rfcomm0"	seems JSR-82 phones is more usable now
-#define DEFAULT_DEVICE  "bluetooth:19"
+//#define DEFAULT_DEVICE  "bluetooth:19"
+#define DEFAULT_DEVICE  "bluetooth:19,tcp:5197,web:5080"
 
 #define RFCOMM_DEVICE   "rfcomm"
 #define AT_DEVICE   	"at:"
