@@ -1,8 +1,8 @@
 //
 // anyRemote
-// a bluetooth remote for your PC.
+// a wi-fi or bluetooth remote for your PC.
 //
-// Copyright (C) 2006-2012 Mikhail Fedotov <anyremote@mail.ru>
+// Copyright (C) 2006-2014 Mikhail Fedotov <anyremote@mail.ru>
 //
 // This program is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -29,7 +29,9 @@
 
 #ifdef USE_XTEST
 #include <X11/Xlib.h>
+#include <X11/Xutil.h>
 #include <X11/extensions/XTest.h>
+#include <X11/keysym.h>
 #include <X11/XF86keysym.h>
 #endif
 
@@ -55,6 +57,9 @@ extern char tmp[MAXMAXLEN];
 static Display* disp = NULL;
 #endif
 
+static int modifierCtrl = 0;
+static int modifierAlt  = 0;
+
 void freeDisplay()
 {
 #ifdef USE_XTEST
@@ -66,6 +71,42 @@ void freeDisplay()
 }
 
 #ifdef USE_XTEST
+
+static void fakeKeyProcess(KeyCode fakeKey, int shift)
+{
+    DEBUG2("fakeKeyProcess Key=%d Shift=%d Ctrl=%d Alt=%d", fakeKey, shift, modifierCtrl, modifierAlt);
+    
+    KeyCode shiftKey;
+    KeyCode altKey;
+    KeyCode ctrlKey;
+    
+    if (shift) {
+        shiftKey = XKeysymToKeycode(disp,XK_Shift_L);
+        XTestFakeKeyEvent(disp, shiftKey, True, CurrentTime);
+    }
+    if (modifierCtrl) {
+        ctrlKey = XKeysymToKeycode(disp,XK_Control_L);
+        XTestFakeKeyEvent(disp, ctrlKey, True, CurrentTime);
+    }
+    if (modifierAlt) {
+        altKey = XKeysymToKeycode(disp,XK_Alt_L);
+        XTestFakeKeyEvent(disp, altKey, True, CurrentTime);
+    }
+
+    XTestFakeKeyEvent(disp, fakeKey, True, CurrentTime);
+    XTestFakeKeyEvent(disp, fakeKey, False, CurrentTime);
+     
+    if (modifierAlt) {
+        XTestFakeKeyEvent(disp, altKey, False, CurrentTime);
+    }
+    if (modifierCtrl) {
+        XTestFakeKeyEvent(disp, ctrlKey, False, CurrentTime);
+    }
+    if (shift) {
+        XTestFakeKeyEvent(disp, shiftKey, False, CurrentTime);
+    }
+}
+
 static void mouseClick(int button)
 {
     XTestFakeButtonEvent(disp, button, True, CurrentTime);
@@ -79,54 +120,123 @@ static void mouseDoubleClick()
     mouseClick(1);
 }
 
-static KeyCode keysymStr2keycode(Display* disp, const char* keysymStr)
+static void setModifier(const char *btn, int on)
 {
-    KeySym ks = XStringToKeysym(keysymStr);
-    if(ks == NoSymbol) {
-
-        if(strcmp("AudioForward",keysymStr) == 0)        	{
-            ks = XF86XK_AudioForward;
-        } else if (strcmp("AudioLowerVolume",keysymStr) == 0)  	{
-            ks = XF86XK_AudioLowerVolume;
-        } else if(strcmp("AudioMute",keysymStr) == 0)	      	{
-            ks = XF86XK_AudioMute;
-        } else if(strcmp("AudioNext",keysymStr) == 0)	      	{
-            ks = XF86XK_AudioNext;
-        } else if(strcmp("AudioPause",keysymStr) == 0)	      	{
-            ks = XF86XK_AudioPause;
-        } else if(strcmp("AudioPlay",keysymStr) == 0)	      	{
-            ks = XF86XK_AudioPlay;
-        } else if(strcmp("AudioPrev",keysymStr) == 0)	      	{
-            ks = XF86XK_AudioPrev;
-        } else if(strcmp("AudioRaiseVolume",keysymStr) == 0)	{
-            ks = XF86XK_AudioRaiseVolume;
-        } else if(strcmp("AudioRewind",keysymStr) == 0)	      	{
-            ks = XF86XK_AudioRewind;
-        } else if(strcmp("Back",keysymStr) == 0)   		{
-            ks = XF86XK_Back;
-        } else if(strcmp("Close",keysymStr) == 0)   		{
-            ks = XF86XK_Close;
-        } else if(strcmp("Eject",keysymStr) == 0)   		{
-            ks = XF86XK_Eject;
-        } else if(strcmp("Forward",keysymStr) == 0)   		{
-            ks = XF86XK_Forward;
-        } else if(strcmp("MonBrightnessDown",keysymStr) == 0)   	{
-            ks = XF86XK_MonBrightnessDown;
-        } else if(strcmp("MonBrightnessUp",keysymStr) == 0)     	{
-            ks = XF86XK_MonBrightnessUp;
-        } else if(strcmp("PowerOff",keysymStr) == 0)   		{
-            ks = XF86XK_PowerOff;
-        } else if(strcmp("WakeUp",keysymStr) == 0)   		{
-            ks = XF86XK_WakeUp;
-        } else {
-            DEBUG2("couldn't map keysymbol %s", keysymStr);
-            ks = XStringToKeysym("space");
-        }
+    if (strcmp("Ctrl", btn) == 0) {
+         modifierCtrl = on;
+    } else if (strcmp("Alt", btn) == 0) { 
+        modifierAlt   = on;  
+    } else {
+        DEBUG2("emulateCommands: improper modifier %s", btn);
     }
+}
+
+static struct 
+{ 
+    const char* ksStr;
+    int         shift;
+    KeySym      ks;
+} 
+ksTable[] = {
+    {"!",  1, XK_exclam},
+    {"\"", 0, XK_quotedbl},
+    {"#",  1, XK_numbersign},
+    {"$",  1, XK_dollar},
+    {"%",  1, XK_percent},
+    {"&",  1, XK_ampersand},
+    {"`",  1, XK_apostrophe},
+    {"*",  1, XK_asterisk},
+    {"+",  1, XK_plus},
+    {"-",  0, XK_minus},
+    {".",  0, XK_period},
+    {"/",  0, XK_slash} ,
+    {":",  1, XK_colon},
+    {"<",  1, XK_less},
+    {"=",  0, XK_equal},
+    {">",  1, XK_greater},
+    {"?",  1, XK_question},
+    {"[",  0, XK_bracketleft},
+    {"\\", 0, XK_backslash},
+    {"]",  0, XK_bracketright},
+    {"^",  1, XK_asciicircum},
+    {"_",  1, XK_underscore},
+    {"@",  1, XK_at},
+     
+    {"parenleft",  1, XK_parenleft},
+    {"parenright", 1, XK_parenright},
+    {"comma",      0, XK_comma},
+    {"semicolon",  0, XK_semicolon},
+    {"space",      0, XK_space},
+   
+    {"Back",             0, XF86XK_Back},
+    {"Close",            0, XF86XK_Close},
+    {"Eject",            0, XF86XK_Eject},
+    {"Forward",          0, XF86XK_Forward},
+    
+    {"AudioForward",     0, XF86XK_AudioForward},
+    {"AudioLowerVolume", 0, XF86XK_AudioLowerVolume},
+    {"AudioMute",        0, XF86XK_AudioMute},
+    {"AudioNext",        0, XF86XK_AudioNext},
+    {"AudioPause",       0, XF86XK_AudioPause}, 
+    {"AudioPlay",        0, XF86XK_AudioPlay}, 
+    {"AudioPrev",        0, XF86XK_AudioPrev},
+    {"AudioRaiseVolume", 0, XF86XK_AudioRaiseVolume},
+    {"AudioRewind",      0, XF86XK_AudioRewind},
+    
+    {"MonBrightnessDown",0, XF86XK_MonBrightnessDown},
+    {"MonBrightnessUp",  0, XF86XK_MonBrightnessUp},
+    
+    {"PowerOff",         0, XF86XK_PowerOff},
+    {"WakeUp",           0, XF86XK_WakeUp},
+    { NULL,              0, NoSymbol}
+};
+
+static KeyCode keysymStr2keycode(Display* disp, const char* keysymStr, int* shift)
+{
+    (*shift) = 0;
+    KeySym ks = NoSymbol;
+    
+    int i=0;
+    while (ksTable[i].ksStr) {
+        if(strcmp(ksTable[i].ksStr,keysymStr) == 0) {
+            ks = ksTable[i].ks;
+            (*shift) = ksTable[i].shift;
+            DEBUG2("match keysymbol %s %ld %d", keysymStr,ks,(*shift));
+            break;
+        }
+        i++;
+    }
+    
+    if(ks == NoSymbol) {
+       ks = XStringToKeysym(keysymStr);
+    }
+    DEBUG2("matched %s %ld %d", keysymStr,ks,(*shift));
+    
+    if (ks == NoSymbol) {
+        DEBUG2("couldn't map keysymbol %s", keysymStr);
+        ks = XStringToKeysym("space");
+        (*shift) = 0;
+    }
+    
     KeyCode kk = XKeysymToKeycode(disp, ks);
     if (kk == 0) {
         kk = XKeysymToKeycode(disp,XStringToKeysym("space"));
     }
+    
+    // check shifted
+    KeySym ks2 = XKeycodeToKeysym(disp, kk, 0);
+    if (ks2 != NoSymbol) {
+        if (ks2 != ks) {  // possibly shifted
+            KeySym lower_return;
+            KeySym upper_return;
+            XConvertCase(ks2, &lower_return, &upper_return);
+            
+            if (upper_return == ks) {
+                (*shift) = 1;
+            }
+        }
+    }
+    
     return kk;
 }
 
@@ -154,6 +264,7 @@ int emulateCommands(int subtype, const char *descr, const char* cmdString, cmdPa
     char * cmd = strdup(descr);
 
     KeyCode fakeKey;
+    int shift = 0;
 
     char* token = strtok(cmd,",");
     while (token) {
@@ -174,7 +285,7 @@ int emulateCommands(int subtype, const char *descr, const char* cmdString, cmdPa
                         continue;
                     }
                 } else {
-                    fakeKey = keysymStr2keycode(disp,key);
+                    fakeKey = keysymStr2keycode(disp,key,&shift);
                 }
                 XTestFakeKeyEvent(disp, fakeKey, False, CurrentTime);
             } else {
@@ -196,7 +307,7 @@ int emulateCommands(int subtype, const char *descr, const char* cmdString, cmdPa
                         continue;
                     }
                 } else {
-                    fakeKey = keysymStr2keycode(disp,key);
+                    fakeKey = keysymStr2keycode(disp,key,&shift);
                 }
                 XTestFakeKeyEvent(disp, fakeKey, True, CurrentTime );
             } else {
@@ -204,12 +315,19 @@ int emulateCommands(int subtype, const char *descr, const char* cmdString, cmdPa
             }
 
         } else if (strcmp(token,"key") == 0) {
+
             char * key = strtok(NULL,",");
+            
+            //DEBUG2("emulateCommands: PARSE %s %c >%s<", token, *(token+4), key ? key : "NULL");
+            
             if (key) {
                 if (strcmp(key,"by_value") == 0) {
                     char * keyval = strtok(NULL,",");
                     if (keyval) {
                         //printf("Emulate key %s\n",keyval);
+                        
+                        DEBUG2("emulateCommands: key value is %s", keyval);
+                        
                         fakeKey = XKeysymToKeycode(disp,strtol(keyval, (char**) NULL, 0));
                         if (fakeKey == 0) {
                             logger(L_DBG, "emulateCommands: improper key value");
@@ -220,14 +338,36 @@ int emulateCommands(int subtype, const char *descr, const char* cmdString, cmdPa
                         continue;
                     }
                 } else {
-                    fakeKey = keysymStr2keycode(disp,key);
+                
+                    DEBUG2("emulateCommands: key is %s", key);
+                    fakeKey = keysymStr2keycode(disp,key,&shift);
                 }
-                //printf("Emulate key dec=%d hex=0x%x\n",fakeKey,fakeKey);
-                XTestFakeKeyEvent(disp, fakeKey, True, CurrentTime);
-                XTestFakeKeyEvent(disp, fakeKey, False, CurrentTime);
+                
+                fakeKeyProcess(fakeKey, shift);
+                
             } else {
                 logger(L_DBG, "emulateCommands: improper key");
             }
+            
+        } else if (strcmp(token,"modifier") == 0) {
+            
+            char * on = strtok(NULL,",");
+            if (on) {
+                int v = 0;
+                if (strcmp("1", on) == 0) {
+                    v = 1;
+                }
+                
+                char * mval = strtok(NULL,",");
+                if (mval) {
+                    setModifier(mval, v);
+                } else {
+                    logger(L_DBG, "emulateCommands: modifier absent");
+                }
+            } else {
+                logger(L_DBG, "emulateCommands: improper modifier value");
+            }
+
 
         } else if (strcmp(token,"mousemove") == 0) {
             char * x = strtok(NULL,",");
@@ -298,3 +438,4 @@ int emulateCommands(int subtype, const char *descr, const char* cmdString, cmdPa
 
     return EXIT_OK;
 }
+
